#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <SDL/SDL.h>
#include "pt_header.h"
#include "pt_helpers.h"
#include "pt_blep.h"
#include "pt_config.h"

static int PAN1 = 96, PAN2 = 160, PAN3 = 160, PAN4 = 96;

#define DENORMAL_OFFSET 1E-10f

typedef struct paula_filter_state
{
    float LED[4];
    float high[2];
} Filter;

typedef struct paula_filter_coefficients
{
    float LED;
    float LEDFb;
    float high;
} FilterC;

typedef struct voice_data
{
    const signed char *data;
    float vol, panL, panR;
    int length, loopLength, loopEnd;

    const signed char *newData;
    int newLength, newLoopLength, newLoopEnd;
    int swapSampleFlag;

    float frac, rate;
    int index;
} Voice;

typedef struct scope_data
{
    signed char dataA[PAULA_CHANNELS * 40];
    signed char dataB[PAULA_CHANNELS * 40];
    signed char *working;
    signed char *scopeData;
    signed char scopeDirty;
} Scope;

static Voice v[PAULA_CHANNELS];
Scope vs;
static Filter filter;
static FilterC filterC;

static float *mixBufferL, *mixBufferR;

static float vsyncBlockLength = 0;
static float vsyncSamplesLeft = 0;

static BLEP blep[PAULA_CHANNELS];
static BLEP blepVol[PAULA_CHANNELS];

int processSamples(int numSamples); // defined in pt_modplayer.c

#ifdef _MSC_VER
inline static int ftol(const float f)
{
    __asm 
    {
        fld f
        push eax
        fistp dword ptr [esp]
        pop eax
    }
}
#else
#define ftol(f) (signed int)(f)
#endif

int getChReadPos(int ch)
{
    return v[ch].data ? v[ch].index : -1;
}

static float calcRcCoeff(float sampleRate, float cutOffFreq)
{
    if (cutOffFreq >= (sampleRate / 2.0f))
    {
        return 1.0f;
    }

    return ((2.0f * 3.141592f) * cutOffFreq / sampleRate);
}

void mixerSwapChSource(int ch, const signed char *src, int length, int loopStart, int loopLength)
{
    // Emulate a hilarious and not-well-known
    // paula bug/pt behavior (not sure) ...
    // If you swap sample in-realtime in PT 1.x
    // without putting a period, the current
    // sample will play through first (either
    // >len or >loop_len), then when that is
    // reached you change to the new sample you
    // put earlier.

    v[ch].swapSampleFlag = true;
    v[ch].newData = src;
    v[ch].newLength = length;
    v[ch].newLoopLength = loopLength;
    v[ch].newLoopEnd = loopLength + loopStart;
}

void mixerSetChSource(int ch, const signed char *src, int length, int loopStart, int loopLength, int offset)
{
    v[ch].swapSampleFlag = false;
    v[ch].data = src;
    v[ch].length = length;
    v[ch].index = offset;
    v[ch].frac = 0.0f;
    v[ch].loopEnd = loopStart + loopLength;
    v[ch].loopLength = loopLength;

    if (v[ch].index > 0)
    {
        if (v[ch].loopLength >= 4)
        {
            if (v[ch].index >= v[ch].loopEnd)
            {
                v[ch].index = 0;
            }
        }
        else if (v[ch].index >= v[ch].length)
        {
            v[ch].data = NULL;
        }
    }
}

static void mixerSetChPan(int ch, int pan)
{
    float p = (float)pan;
    v[ch].panL = 256.0f - p;
    v[ch].panR = p;
}

void mixerCutChannels(void)
{
    memset(v, 0, sizeof (v));
    memset(&vs, 0, sizeof (vs));
    memset(blep, 0, sizeof (blep));
    memset(blepVol, 0, sizeof (blepVol));
    memset(&filter, 0, sizeof (filter));

    vs.working = vs.dataA;
    vs.scopeData = vs.dataB;
    vsyncSamplesLeft = 0.0f;

    mixerSetChPan(0, PAN1);
    mixerSetChPan(1, PAN2);
    mixerSetChPan(2, PAN3);
    mixerSetChPan(3, PAN4);
}

void mixerSetChVol(int ch, float vol)
{
    v[ch].vol = vol;
}

void mixerSetChRate(int ch, float rate)
{
    v[ch].rate = rate;
}

static void outputAudio(signed short *target, int numSamples)
{
    int i, j;
    signed short *out;

    memset(mixBufferL, 0, numSamples * sizeof (float));
    memset(mixBufferR, 0, numSamples * sizeof (float));

    for (i = 0; i < PAULA_CHANNELS; ++i)
    {
        float mutedVol = -1.0f;
        j = 0;

        if (editor.muted[i] == true)
        {
            mutedVol = v[i].vol;
            v[i].vol = 0.0f;
        }

        if (v[i].data != NULL)
        {
            for (j = 0; j < numSamples; ++j)
            {
                float tempSample = v[i].data ? (float)v[i].data[v[i].index] : 0.0f;
                float tempVolume = v[i].data ? v[i].vol : 0.0f;

                // if sample data changes, anti-alias the step
                // this code assumes index can't change > 1 per sample
                if (tempSample != blep[i].lastValue)
                {
                    float delta = blep[i].lastValue - tempSample;
                    blep[i].lastValue = tempSample;
                    blepAdd(&blep[i], v[i].frac / v[i].rate, delta);
                }

                // if volume data changes, anti-alias the step
                if (tempVolume != blepVol[i].lastValue)
                {
                    float delta = blepVol[i].lastValue - tempVolume;
                    blepVol[i].lastValue = tempVolume;
                    blepAdd(&blepVol[i], 0.0f, delta); // TODO: will insert tick fraction here
                }

                // add sample anti-alias data to sample value and scale by volume
                if (blep[i].samplesLeft > 0)
                {
                    tempSample += blepRun(&blep[i]);
                }

                // add volume anti-alias pulse data to volume value
                if (blepVol[i].samplesLeft > 0)
                {
                    tempVolume += blepRun(&blepVol[i]);
                }

                tempSample *= tempVolume;

                mixBufferL[j] += (tempSample * v[i].panL);
                mixBufferR[j] += (tempSample * v[i].panR);

                if (v[i].data != NULL)
                {
                    v[i].frac += v[i].rate;

                    if (v[i].frac >= 1.0f)
                    {
                        v[i].index++;
                        v[i].frac -= 1.0f;

                        if (v[i].loopLength >= 4)
                        {
                            if (v[i].index >= v[i].loopEnd)
                            {
                                if (v[i].swapSampleFlag == true)
                                {
                                    v[i].swapSampleFlag = false;

                                    if (v[i].newLoopLength <= 2)
                                    {
                                        v[i].data = NULL;

                                        break;
                                    }

                                    v[i].data = v[i].newData;
                                    v[i].length = v[i].newLength;
                                    v[i].loopEnd = v[i].newLoopEnd;
                                    v[i].loopLength = v[i].newLoopLength;
                                    v[i].index = v[i].loopEnd - v[i].loopLength;
                                }
                                else
                                {
                                    v[i].index -= v[i].loopLength;
                                }
                            }
                        }
                        else if (v[i].index >= v[i].length)
                        {
                            j++;

                            tempSample = 0.0f;

                            // if the sample has ended, insert the impulse for the last to-zero step
                            if (tempSample != blep[i].lastValue)
                            {
                                float delta = blep[i].lastValue - tempSample;
                                blep[i].lastValue = tempSample;
                                blepAdd(&blep[i], v[i].frac / v[i].rate, delta);
                            }

                            if (v[i].swapSampleFlag == true)
                            {
                                v[i].swapSampleFlag = false;

                                v[i].data = v[i].newData;
                                v[i].length = v[i].newLength;
                                v[i].loopEnd = v[i].newLoopEnd;
                                v[i].loopLength = v[i].newLoopLength;
                            }
                            else
                            {
                                v[i].data = NULL;

                                break;
                            }
                        }
                    }
                }
            }
        }
        else if (v[i].swapSampleFlag == true)
        {
            v[i].swapSampleFlag = false;

            v[i].data = v[i].newData;
            v[i].length = v[i].newLength;
            v[i].loopEnd = v[i].newLoopEnd;
            v[i].loopLength = v[i].newLoopLength;
        }

        // sometimes a channel might be empty, 
        // but the blep-buffer still contains data.
        if ((j < numSamples) && (v[i].data == NULL)
            && ((blep[i].samplesLeft > 0) || (blepVol[i].samplesLeft > 0)))
        {
            for (; j < numSamples; ++j)
            {
                float tempSample = blep[i].lastValue;
                float tempVolume = blepVol[i].lastValue;
                
                if (blep[i].samplesLeft > 0)
                {
                    tempSample += blepRun(&blep[i]);
                }

                if (blepVol[i].samplesLeft > 0)
                {
                    tempVolume += blepRun(&blepVol[i]);
                }

                tempSample *= tempVolume;

                mixBufferL[j] += (tempSample * v[i].panL);
                mixBufferR[j] += (tempSample * v[i].panR);
            }
        }

        if (mutedVol != -1.0f)
        {
            v[i].vol = mutedVol;
        }
    }

    out = target;
    for (i = 0; i < numSamples; ++i)
    {
        float L = mixBufferL[i];
        float R = mixBufferR[i];

        if (editor.useLEDFilter == true)
        {
            // sallen-key 12dB/o lowpass with Q ~= 1/sqrt(2) (butterworth)
            filter.LED[0] += (filterC.LED * (L - filter.LED[0])
                + filterC.LEDFb * (filter.LED[0] - filter.LED[1]) + DENORMAL_OFFSET);
            filter.LED[1] += (filterC.LED * (filter.LED[0] - filter.LED[1]) + DENORMAL_OFFSET);
            filter.LED[2] += (filterC.LED * (R - filter.LED[2])
                + filterC.LEDFb * (filter.LED[2] - filter.LED[3]) + DENORMAL_OFFSET);
            filter.LED[3] += (filterC.LED * (filter.LED[2] - filter.LED[3]) + DENORMAL_OFFSET);

            L = filter.LED[1];
            R = filter.LED[3];
        }

        // for blocking DC
        L -= filter.high[0];
        R -= filter.high[1];

        // update filter
        filter.high[0] += (filterC.high * L + DENORMAL_OFFSET);
        filter.high[1] += (filterC.high * R + DENORMAL_OFFSET);

        // normalize volume
        L *= (1.0f / 96.0f);
        R *= (1.0f / 96.0f);

        // clamp in case of overflow
        L = CLAMP(L, -32768.0f, 32767.0f);
        R = CLAMP(R, -32768.0f, 32767.0f);

        // convert to int and send to output
        *out++ = ftol(L);
        *out++ = ftol(R);
    }
}

// run fake voices from the current offset and rate to fill the scopes
void fill_scope_block(void)
{
    if (vs.scopeDirty == true)
    {
        int i, j;
        for (i = 0; i < PAULA_CHANNELS; ++i)
        {
            const signed char *data = v[i].data;
            signed char *scope = vs.working + (i * 40);

            if (data != NULL)
            {
                int index = v[i].index;
                int loopEnd = v[i].loopEnd;
                int loopLength = v[i].loopLength;
                int length = v[i].length;
                float vol = v[i].vol * (1.0f / 512.0f);

                for (j = 0; j < 40; ++j)
                {
                    scope[j] = ftol(floorf((data[index] * vol) + 0.5f));

                    index++;

                    if (loopLength >= 4)
                    {
                        if (index >= loopEnd)
                        {
                            index -= loopLength;
                        }
                    }
                    else if (index >= length)
                    {
                        memset(scope + j, 0, 40 - j);

                        break;
                    }
                }
            }
            else
            {
                memset(scope, 0, 40);
            }
        }

        // flip the buffers
        vs.scopeDirty = false;
        vs.scopeData = vs.working;
        vs.working = (vs.working == vs.dataA) ? vs.dataB : vs.dataA;
    }

    // if scope wasn't dirty, we dropped a frame
}

void audioCallback(void *userdata, unsigned char *stream, int len)
{
    signed short *out = (signed short *)stream;

    int samplesLeft = len / 4;
    while (samplesLeft)
    {
        int igen, gen;
        
        gen = processSamples(samplesLeft);
        samplesLeft -= gen;

        while (gen)
        {
            if (gen >= vsyncSamplesLeft)
            {
                igen = ftol(floorf(vsyncSamplesLeft));
                gen -= igen;
                vsyncSamplesLeft -= igen;
            }
            else
            {
                igen = gen;
                vsyncSamplesLeft -= gen;
                gen = 0;
            }

            vsyncSamplesLeft = 0.0f;

            outputAudio(out, igen);
            out += (igen * 2);

            if (vsyncSamplesLeft <= 1.0f)
            {
                vsyncSamplesLeft += vsyncBlockLength;
                fill_scope_block();
            }
        }
    }
}

static int mixerOpen(int soundFrequency, int soundBufferSize, int baseClockRegion)
{
    SDL_AudioSpec fmt;

    fmt.freq = soundFrequency;
    fmt.format = AUDIO_S16SYS;
    fmt.channels = 2;
    fmt.samples = (short)soundBufferSize;
    fmt.callback = audioCallback;
    fmt.userdata = NULL;

    vsyncBlockLength = (float)soundFrequency / 50.0f;

    if (SDL_OpenAudio(&fmt, NULL) < 0)
    {
        fprintf(stderr, "Unable to open audio: %s\n", SDL_GetError());

        return 0;
    }

    mixBufferL = (float *)calloc(soundBufferSize, sizeof (float));
    if (mixBufferL == NULL)
    {
        fprintf(stderr, "Unable to open audio: Could not allocate sound buffers\n");

        return 0;
    }

    mixBufferR = (float *)calloc(soundBufferSize, sizeof (float));
    if (mixBufferR == NULL)
    {
        fprintf(stderr, "Unable to open audio: Could not allocate sound buffers\n");

        return 0;
    }

    createFrequencyTable(soundFrequency, baseClockRegion);

    // Amiga filter emulation
    //
    // First comes a static lowpass 6dB formed by the supply current
    // from the Paula's mixture of channels A+B / C+D into the opamp with
    // 100nF capacitor and 360 ohm resistor feedback in inverting mode biased by
    // dac vRef (used to center the output).
    //
    // Under spice simulation the circuit yields (F - 3dB) = 4400Hz.
    // In the A1200, the cutoff is 26000Hz, so the static filter is disabled.
    //
    // Next comes a bog-standard Sallen-Key filter with:
    // R = 10k ohm
    // C0 = 6.8nF
    // C1 = 3.9nF
    // Q ~= 1/sqrt(2)
    //
    // This filter is optionally bypassed by a MPF-102 JFET when
    // the LED is disabled.
    //
    // Under spice simulation the circuit yields (F - 3dB) = 2800Hz.
    // 90 phase = 3000Hz (so, should oscillate at 3kHz!)
    //
    // The buffered output of the Sallen-Key passes into an RC highpass with:
    // C = 22F (also C = 330nF, for improved high-frequency)
    // R = 1.39k ohm
    //
    // F = 1 / 2piRC
    // High Hz = 5.2~ = 1 / (2pi * 0.000022 * 1390)
    //
    // Under spice simulation the circuit yields (F - 3dB) = 5.2Hz.

    filterC.LED = calcRcCoeff((float)soundFrequency, 3000.0f);
    filterC.LEDFb = 0.125f + 0.125f / (1.0f - filterC.LED);
    // Fb = 0.125 : Q ~= 1/sqrt(2) (Butterworth)

    // ~30Hz is used here rather than the 5.2Hz value in order to
    // lower the total amount of clipping taking place, and for rumble.
    //
    // In the Amiga circuit there is about 20dB (10x) of headroom available,
    // here we have none at all. Many modules use samples with significant DC
    // offsets and these will inevitably lead to clipping with a low Hz value here.
    //
    // Another solution would be to provide more headroom (12dB - 4x),
    // however that would require decreasing the output gain by a proportionate amount.
    //
    // Output with 30Hz is closer to what might be received in a real world situation anyway.
    // Some low-frequency content will be lost due to this.

    filterC.high = calcRcCoeff((float)soundFrequency, 30.0f);

    // Empirical testing shows these values to be accurate, in terms of coefficients.
    // Need to test against the output of a real A500 / A1200 Paula to be certain of match.

    editor.useLEDFilter = false;

    return 1;
}

static void mixerClose(void)
{
    SDL_CloseAudio();

    if (mixBufferL != NULL)
    {
        free(mixBufferL);
    }

    if (mixBufferR != NULL)
    {
        free(mixBufferR);
    }
}

void mixerCalcChPans(int stereoSeparation)
{
    int scaledPanPos = (stereoSeparation * 128) / 100;

    PAN1 = 128 - scaledPanPos;
    PAN2 = 128 + scaledPanPos;
    PAN3 = 128 + scaledPanPos;
    PAN4 = 128 - scaledPanPos;

    mixerSetChPan(0, PAN1);
    mixerSetChPan(1, PAN2);
    mixerSetChPan(2, PAN3);
    mixerSetChPan(3, PAN4);
}

int audioOpen(const PTCONFIG *config)
{
    mixerCutChannels();
    mixerCalcChPans(config->stereoSeparation);

    if (!mixerOpen(config->soundFrequency, config->soundBufferSize, config->region))
    {
        return false;
    }

    SDL_PauseAudio(false);

    return true;
}

void audioClose(void)
{
    editor.songPlaying = false;

    mixerCutChannels();
    SDL_PauseAudio(true);

    mixerClose();
}
