#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h> // for toupper()
#include <SDL/SDL.h>

#ifdef WIN32

#ifdef _MSC_VER
#pragma warning(disable: 4996) // for deprecated  POSIX warnings
#endif

#include <io.h>
#else
#include <unistd.h>
#endif

#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>
#include <limits.h>
#include "pt_header.h"
#include "pt_dirent.h"
#include "pt_textout.h"
#include "pt_diskop.h"

#define DISKOP_LIST_SIZE 8

static int diskOpNumFiles = 0;

static int diskOpEntryType[DISKOP_LIST_SIZE];
static char diskOpEntryDate[DISKOP_LIST_SIZE][7];
static char diskOpEntrySize[DISKOP_LIST_SIZE][8];
static char diskOpEntryPath[DISKOP_LIST_SIZE][PATH_MAX];
static char diskOpEntry[DISKOP_LIST_SIZE][PATH_MAX];

static int diskOpFillBuffer(const char *path, int fileEntryStart);

int diskOpEntryIsDir(int index)
{
    if (index < DISKOP_LIST_SIZE)
    {
        return (diskOpEntryType[index] == DISKOP_DIR);
    }

    return -1;
}

char *diskOpGetEntry(int index)
{
    if (index < DISKOP_LIST_SIZE)
    {
        if (diskOpEntry[index][0] != '\0')
        {
            return diskOpEntry[index];
        }
    }

    return NULL;
}

char *diskOpGetEntryPath(int index)
{
    if (index < DISKOP_LIST_SIZE)
    {
        if (diskOpEntryPath[index][0] != '\0')
        {
            return diskOpEntryPath[index];
        }
    }

    return NULL;
}

void diskOpGoParent(void)
{
    char tmpDirBuf[PATH_MAX];
#ifdef WIN32
    sprintf(tmpDirBuf, "%s..", editor.diskop.currentPath);
#else
    sprintf(tmpDirBuf, "%s/..", editor.diskop.currentPath);
#endif
    diskOpSetPath(tmpDirBuf);
}

void diskOpSetPath(const char *path)
{
    int pathLen = strlen(path);
    if (pathLen < PATH_MAX)
    {
        strcpy(editor.diskop.currentPath, path);

#ifdef WIN32
        if (editor.diskop.currentPath[pathLen - 1] == '*')
        {
            editor.diskop.currentPath[pathLen - 1] = '\0';
        }
#endif
    }
}

void diskOpDraw(unsigned int *frameBuffer)
{
    int i;

    char diskOpTemp[PATH_MAX];

    editor.diskop.numFiles = diskOpNumFiles - DISKOP_LIST_SIZE;

    // print out dir path once, clamp with ".." if long string
    if (strlen(editor.diskop.currentPath) > 18)
    {
        strncpy(diskOpTemp, editor.diskop.currentPath, 18);

        diskOpTemp[18] = '\0';
        diskOpTemp[17] = '.';
        diskOpTemp[16] = '.';

        textOut(frameBuffer, 32, 36, diskOpTemp, 0x00000000);
    }
    else
    {
        textOut(frameBuffer, 32, 36, editor.diskop.currentPath, 0x00000000);
    }

    // if needed, update the file list and add entries
    if (editor.diskop.forceUpdate == true)
    {
        editor.diskop.forceUpdate = false;

        if (diskOpNumFiles > DISKOP_LIST_SIZE)
        {
            if (editor.diskop.scrollOffset > (diskOpNumFiles - DISKOP_LIST_SIZE))
            {
                editor.diskop.scrollOffset = (diskOpNumFiles - DISKOP_LIST_SIZE);
            }

            diskOpFillBuffer(editor.diskop.currentPath, editor.diskop.scrollOffset);
        }
        else
        {
            diskOpFillBuffer(editor.diskop.currentPath, 0);
        }
    }

    // print out file entries, clamp with ".." if long string
    for (i = 0; (i < diskOpNumFiles) && (i < DISKOP_LIST_SIZE); ++i)
    {
        int charYPad = i * 6;

        if (strlen(diskOpEntry[i]) > 23)
        {
            strncpy(diskOpTemp, diskOpEntry[i], 23);

            diskOpTemp[23] = '\0';
            diskOpTemp[22] = '.';
            diskOpTemp[21] = '.';

            textOut(frameBuffer, 64, 46 + charYPad, diskOpTemp, 0x00FFDD00);
        }
        else
        {
            textOut(frameBuffer, 64, 46 + charYPad, diskOpEntry[i], 0x00FFDD00);
        }

        textOut(frameBuffer, 8, 46 + charYPad, diskOpEntryDate[i], 0x00FFDD00);
        textOut(frameBuffer, 248, 46 + charYPad, diskOpEntrySize[i], 0x00FFDD00);
    }
}

static int diskOpFillBuffer(const char *path, int fileEntryStart)
{
    int i;

    int fileNameLength;
    int fileHandle, fileIndex, fileOffset;
    char fileExtFirst[4], fileExtLast[4];

    DIR *dir;
    struct dirent *ent;
    struct stat fileStat;

    dir = opendir(path);
    if (dir == NULL)
    {
        return false;
    }

    {
#ifdef WIN32
        diskOpSetPath(dir->patt);
#else
        char tempPath[PATH_MAX];
        getcwd(tempPath, PATH_MAX);
        if (*tempPath == '.')
        {
            *tempPath = '/';
        }

        diskOpSetPath(tempPath);
#endif
    }

    // clear lists first
    for (i = 0; i < DISKOP_LIST_SIZE; ++i)
    {
        memset(diskOpEntryPath[i], '\0', PATH_MAX);
        memset(diskOpEntry[i], '\0', PATH_MAX);
        memset(diskOpEntrySize[i], '\0', 8);
        memset(diskOpEntryDate[i], '\0', 7);
    }

    diskOpNumFiles = 0;
    fileIndex = 0;
    fileOffset = 0;

    // read directory and scan files
    while ((ent = readdir(dir)) != NULL)
    {
        fileNameLength = strlen(ent->d_name);

        // MOD.<filename>
        fileExtFirst[0] = (char)toupper(ent->d_name[0]);
        fileExtFirst[1] = (char)toupper(ent->d_name[1]);
        fileExtFirst[2] = (char)toupper(ent->d_name[2]);
        fileExtFirst[3] = (char)toupper(ent->d_name[3]);

        // <filename>.MOD
        fileExtLast[0] = (char)toupper(ent->d_name[fileNameLength - 4]);
        fileExtLast[1] = (char)toupper(ent->d_name[fileNameLength - 3]);
        fileExtLast[2] = (char)toupper(ent->d_name[fileNameLength - 2]);
        fileExtLast[3] = (char)toupper(ent->d_name[fileNameLength - 1]);

        // a MOD., .MOD or a dir?
        if ((!strncmp(fileExtFirst, "MOD.", 4) || !strncmp(fileExtLast, ".MOD", 4))
            || (!strncmp(fileExtFirst, "STK.", 4) || !strncmp(fileExtLast, ".STK", 4))
            || ((ent->d_type == DT_DIR) && (strcmp(ent->d_name, "."))))
        {
            diskOpNumFiles++;

            // make sure we have rows left to fill
            if (fileIndex < DISKOP_LIST_SIZE)
            {
                // start filling now or later? Used for scrolling in the disk op box
                if (fileOffset >= fileEntryStart)
                {
                    strcpy(diskOpEntry[fileIndex], ent->d_name);

                    // get file size and modification date
                    if (ent->d_type == DT_REG)
                    {
#ifdef WIN32
                        sprintf(diskOpEntryPath[fileIndex], "%s%s", editor.diskop.currentPath, ent->d_name);
#else
                        sprintf(diskOpEntryPath[fileIndex], "%s/%s", editor.diskop.currentPath, ent->d_name);
#endif
                        fileHandle = open(diskOpEntryPath[fileIndex], O_RDONLY);
                        if (fileHandle != -1)
                        {
                            fstat(fileHandle, &fileStat);

                            if (fileStat.st_size > 9999999)
                            {
                                fileStat.st_size = 9999999;
                            }

                            diskOpEntryType[fileIndex] = DISKOP_FILE;
                            sprintf(diskOpEntrySize[fileIndex], "%7d", (int)fileStat.st_size);
                            strftime(diskOpEntryDate[fileIndex], 7, "%d%m%y", localtime(&fileStat.st_mtime));

                            close(fileHandle);
                        }
                    }
                    else if (ent->d_type == DT_DIR)
                    {
                        diskOpEntryType[fileIndex] = DISKOP_DIR;
#ifdef WIN32
                        sprintf(diskOpEntryPath[fileIndex], "%s%s", editor.diskop.currentPath, ent->d_name);
#else
                        sprintf(diskOpEntryPath[fileIndex], "%s/%s", editor.diskop.currentPath, ent->d_name);
#endif
                        strcpy(diskOpEntrySize[fileIndex], "  (DIR)");
                        strcpy(diskOpEntryDate[fileIndex], "      ");
                    }

                    fileIndex++;
                }
                else
                {
                    fileOffset++;
                }
            }
        }
    }

    closedir(dir);

    return true;
}
