#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <SDL/SDL.h>
#include <SDL/SDL_thread.h>
#ifdef WIN32
#include <windows.h>
#endif
#include "pt_header.h"
#include "pt_audio.h"
#include "pt_textout.h"
#include "pt_patternviewer.h"
#include "pt_edit.h"
#include "pt_helpers.h"
#include "pt_buttons.h"
#include "pt_visuals.h"
#include "pt_palette.h"
#include "pt_tables.h"
#include "pt_mouse.h"
#include "pt_diskop.h"
#include "pt_sampler.h"
#include "pt_config.h"
#include "pt_intro.h"
#include "pt_modloader.h"
#include "pt_gfx_icon.h"
#include "pt_gfx_tracker.h"
#include "pt_gfx_clear.h"
#include "pt_gfx_sampler.h"
#include "pt_gfx_spectrum.h"
#include "pt_gfx_diskop.h"
#include "pt_gfx_quit_box.h"
#include "pt_gfx_pointer.h"
#include "pt_gfx_patcursor.h"
#include "pt_gfx_mute_buttons.h"
#include "pt_gfx_vumeter.h"

static const char fineTuneChar1[16] =
{
    ' ', '+', '+', '+', '+', '+', '+', '+',
    '-', '-', '-', '-', '-', '-', '-', '-'
};

static const char fineTuneChar2[16] =
{
    '0', '1', '2', '3', '4', '5', '6', '7',
    '8', '7', '6', '5', '4', '3', '2', '1'
};

static int programRunning = true;

static unsigned int ticker60HzNext = 0;
static unsigned int ticker60HzInterval = 0;
static unsigned int ticker50HzNext = 0;
static unsigned int ticker50HzInterval = 0;

static MODULE *mod = NULL;

MODULE *createNewMod(void);

#ifdef _MSC_VER
inline static int ftol(const float f)
{
    __asm
    {
        fld f
        push eax
        fistp dword ptr [esp]
        pop eax
    }
}
#else
#define ftol(f) (signed int)(f)
#endif

void setMainModSource(MODULE *modEntry)
{
    mod = modEntry;
}

int thread50HzFunc(void *unused)
{
    unsigned int ticker50HzNow, ticker50HzWhen;
    int ticker50HzDelay;

    ticker50HzInterval = 1000 / 50;
    ticker50HzNext = SDL_GetTicks() + ticker50HzInterval;

    while (programRunning == true)
    {
        int i;
        for (i = 0; i < 23; ++i)
        {
            // Sink the four vumeters
            if (i < 4)
            {
                if (editor.vuMeterVolumes[i] > 0)
                {
                    editor.vuMeterVolumes[i]--;
                }
            }

            // Sink the 23 spectrum analyzer bars
            if (editor.spectrumVolumes[i] > 0)
            {
                editor.spectrumVolumes[i]--;
            }
        }

        // Attempt to sync thread's delta to 50Hz
        {
            ticker50HzNow = SDL_GetTicks();
            ticker50HzDelay = ticker50HzNext - ticker50HzNow;

            if (ticker50HzDelay > 0)
            {
                SDL_Delay(ticker50HzDelay);
                ticker50HzNow = SDL_GetTicks();
            }

            ticker50HzWhen = (ticker50HzNow + ticker50HzInterval) / 2;

            ticker50HzNext += ticker50HzInterval;
            if (ticker50HzNext < ticker50HzWhen)
            {
                ticker50HzNext = ticker50HzWhen;
            }
        }
    }

    return 0;
}

int main(int argc, char *argv[])
{
    SDL_Surface *screen, *tempScreenSurface;
    SDL_Surface *iconSurface;
    SDL_Event inputEvent;
    SDL_Thread *thread50Hz;

    PTCONFIG config; 
    MODULE_SAMPLE *currentSample;

    int i, j, k;

    unsigned int ticker60HzNow, ticker60HzWhen;
    int ticker60HzDelay;

    unsigned int *frameBuffer, *tempScreenBuffer;
    unsigned int *frameBufferPointer, *frameBufferPointerX;

    int x, y;
    int dstY, srcY, srcX, xPos, yPos;
    unsigned int pixelOut, paletteIndex;

    signed char *scopePointer, *scopeData;

    int volPeak;

    const unsigned int *sourceBuffer;
    char tempChar;

    extern Scope vs;

    SDL_Rect tempScreenRect =
        { 40, 33, FULL_SCREEN_W - (40 * 2), FULL_SCREEN_H - (33 * 2) };

    SDL_Rect frameBufferRect = { 0, 0, SCREEN_W, SCREEN_H };

#ifdef WIN32
    timeBeginPeriod(1); // set timer precision to 1ms
#endif

    if (SDL_Init(SDL_INIT_AUDIO | SDL_INIT_VIDEO) < 0)
    {
        fprintf(stderr, "Couldn't initialize SDL: %s\n", SDL_GetError());

        return 1;
    }

    audio.audioMixerOpen = true;
    editor.songPlaying = false;

    loadConfig(&config);

    // PAL screen resolution
    screen = SDL_SetVideoMode(config.blit2x ? FULL_SCREEN_W : SCREEN_W,
        config.blit2x ? FULL_SCREEN_H : SCREEN_H, 32, SDL_SWSURFACE);
    if (screen == NULL)
    {
        fprintf(stderr, "Couldn't set video mode: %s\n", SDL_GetError());

        SDL_Quit();

        return 1;
    }

    // This thread ticks in 50Hz, it's used to sink the
    // vumeter and spectrumanalyzer in 50Hz just like
    // on a PAL Amiga.
    thread50Hz = SDL_CreateThread(thread50HzFunc, NULL);
    if (thread50Hz == NULL)
    {
        fprintf(stderr, "Couldn't create a 50Hz timer thread: %s\n", SDL_GetError());

        SDL_Quit();

        return 1;
    }

    SDL_WM_SetCaption("ProTracker", NULL);
    SDL_ShowCursor(SDL_DISABLE);

    iconSurface = SDL_CreateRGBSurface(screen->flags, ICON_WIDTH, ICON_HEIGHT, 32,
        screen->format->Rmask, screen->format->Gmask,
        screen->format->Bmask, screen->format->Amask);
    SDL_SetColorKey(iconSurface, SDL_SRCCOLORKEY,
        SDL_MapRGB(iconSurface->format, 0xFF, 0xFF, 0xFF));

    // Copy main icon to a buffer
    for (i = 0; i < (ICON_WIDTH * ICON_HEIGHT); ++i)
    {
        *((unsigned int *)iconSurface->pixels + i) = iconBitmap[i];
    }

    SDL_WM_SetIcon(iconSurface, NULL);

    // ProTracker doesn't use the 256th scanline
    tempScreenSurface = SDL_CreateRGBSurface(screen->flags, SCREEN_W, SCREEN_H, 32,
        screen->format->Rmask, screen->format->Gmask,
        screen->format->Bmask, screen->format->Amask);

    tempScreenBuffer = (unsigned int *)tempScreenSurface->pixels;
    frameBuffer = (unsigned int *)screen->pixels;

    if (audioOpen(&config) == false)
    {
        fprintf(stderr, "Couldn't initialize ProTracker!\n");

        SDL_FreeSurface(tempScreenSurface);
        SDL_Quit();

        return 1;
    }

    editor.playTime = 0;
    editor.diskop.forceUpdate = false;
    editor.diskop.scrollOffset = 0;
    editor.playMode = PLAY_MODE_NORMAL;
    editor.keyOctave = OCTAVE_HIGH;
    editor.currentMode = MODE_IDLE;
    editor.currentSample = 0;
    editor.ui.throwExit = false;
    editor.ui.exitScreenShown = false;
    editor.ui.samplerScreenShown = false;
    editor.ui.clearScreenShown = false;
    editor.ui.diskOpScreenShown = false;
    editor.ui.introScreenShown = true;
    editor.ui.visualizerMode = VISUAL_QUADRASCOPE;
    editor.ui.pointerMode = POINTER_MODE_IDLE;
    editor.ui.previousPointerMode = editor.ui.pointerMode;
    editor.cursor.channel = 0;
    editor.cursor.pos = 0;
    editor.cursor.mode = CURSOR_NOTE;
    editor.muted[0] = false;
    editor.muted[1] = false;
    editor.muted[2] = false;
    editor.muted[3] = false;
    input.keyb.cursorKeyRepeatCounter = 0;
    input.keyb.leftShiftKeyDown = false;
    input.keyb.leftAltKeyDown = false;
    input.keyb.leftControlKeyDown = false;
    input.keyb.cursorKeyPressed = false;
    input.mouse.leftButtonPressed = false;
    input.mouse.rightButtonPressed = false;
    input.mouse.buttonWaiting = false;
    input.mouse.buttonErrorMsgActive = false;
    input.mouse.repeatCounter = 0;
    input.mouse.buttonWaitCounter = 0;
    input.mouse.errorMsgCounter = 0;
    input.mouse.x = SCREEN_W / 2;
    input.mouse.y = SCREEN_H / 2;

    memset(editor.ui.previousStatusMessage, 0, 17);
    memset(editor.ui.statusMessage, 0, 17);
    memset(editor.spectrumVolumes, 0, 23 * sizeof (int));
    memset(editor.vuMeterVolumes, 0, 4 * sizeof (int));

    mod = createNewMod();
    setPlayerModSource(mod);

    modSetTempo(125);
    modSetSpeed(6);

    diskOpSetPath(".");
    setupSampler(mod);

    pointerSetMode(POINTER_MODE_IDLE, true);
    setStatusMessage("ALL RIGHT", true); // Fill prev status
    setStatusMessage("PROTRACKER", false);

    initIntro();

    ticker60HzInterval = ftol((1.0f * 1000.0f) / 60); // 60 as in 60Hz
    ticker60HzNext = SDL_GetTicks() + ticker60HzInterval;

    currentSample = NULL;

    // Load a .MOD from the command arguments if passed
    if ((argc == 2) && (strlen(argv[1]) > 0))
    {
        MODULE *newMod;

        editor.ui.introScreenShown = false;
        pointerSetMode(POINTER_MODE_IDLE, true);
        setStatusMessage("ALL RIGHT", true);

        newMod = modLoad(argv[1]);
        if (newMod != NULL)
        {
            mod->moduleLoaded = false;
            modFree(mod);

            mod = newMod;
            setPlayerModSource(mod);
            setMainModSource(mod);

            modSetSpeed(6);
            modSetTempo(125);
            modSetPos(0, 0);

            mod->moduleLoaded = true;
        }
        else
        {
            input.mouse.buttonErrorMsgActive = true;
            input.mouse.errorMsgCounter = 0;

            setStatusMessage("MOD LOAD ERROR !", false);
            pointerErrorMode();
        }
    }

    while (programRunning == true)
    {
        SDL_GetMouseState(&input.mouse.x, &input.mouse.y);
        if (config.blit2x == true)
        {
            input.mouse.x = CLAMP(input.mouse.x - 40, 0, (FULL_SCREEN_W - (40 * 2)) - 1) / 2;
            input.mouse.y = CLAMP(input.mouse.y - 33, 0, (FULL_SCREEN_H - (33 * 2)) - 1) / 2;
        }
        else
        {
            input.mouse.x = CLAMP(input.mouse.x, 0, SCREEN_W - 1);
            input.mouse.y = CLAMP(input.mouse.y, 0, SCREEN_H - 1);
        }

        while (SDL_PollEvent(&inputEvent) == true)
        {
            if (editor.ui.throwExit == true)
            {
                programRunning = false;
            }

            if (inputEvent.type == SDL_QUIT)
            {
                editor.ui.exitScreenShown = true;

                pointerSetMode(POINTER_MODE_MSG1, false);
                setStatusMessage("REALLY QUIT ?", false);
            }
            else if (inputEvent.type == SDL_KEYUP)
            {
                keyUpHandler(inputEvent.key.keysym.sym);
            }
            else if (inputEvent.type == SDL_KEYDOWN)
            {
                keyDownHandler(inputEvent.key.keysym.sym, screen, mod);
            }
            else if (inputEvent.type == SDL_MOUSEBUTTONUP)
            {
                mouseButtonUpHandler(inputEvent.button.button);
            }
            else if (inputEvent.type == SDL_MOUSEBUTTONDOWN)
            {
                mouseButtonDownHandler(input.mouse.x, input.mouse.y, screen, inputEvent.button.button, mod);

                if (editor.ui.introScreenShown == true)
                {
                    editor.ui.introScreenShown = false;

                    pointerSetMode(POINTER_MODE_IDLE, true);
                    setStatusMessage("ALL RIGHT", true);
                }
            }
        }

        // fetch a pointer to the current sample's sample struct
        currentSample = &mod->samples[editor.currentSample];

        // blit tracker frame to screen
        memcpy(tempScreenBuffer, trackerFrameBitmap, TRACKER_WIDTH * TRACKER_HEIGHT * sizeof (int));

        if (editor.ui.samplerScreenShown == false)
        {
            drawPattern(tempScreenBuffer, mod);

            // DRAW VUMETERS
            for (i = 0; i < PAULA_CHANNELS; ++i)
            {
                volPeak = editor.vuMeterVolumes[i];
                xPos = 55 + (72 * i);

                for (j = volPeak; j > 0; --j)
                {
                    dstY = ((188 - j) * SCREEN_W) + xPos;
                    srcY = j * 10;

                    for (k = 0; k < 10; ++k)
                    {
                        tempScreenBuffer[dstY + k] = vuMeterBitmap[srcY + k];
                    }
                }
            }

            // DRAW PATTERN CURSOR
            xPos = patternCursorXPos[editor.cursor.pos];
            for (y = 0; y < PATCURSOR_HEIGHT; ++y)
            {
                dstY = (188 + y) * SCREEN_W;
                srcY = y * PATCURSOR_WIDTH;

                for (x = 0; x < PATCURSOR_WIDTH; ++x)
                {
                    pixelOut = patternCursorBitmap[srcY + x];
                    if (pixelOut != COLOR_KEY)
                    {
                        tempScreenBuffer[dstY + x + xPos] = pixelOut;
                    }
                }
            }
        }

        if (editor.ui.diskOpScreenShown == false)
        {
            // DRAW QUADRASCOPES OR SPECTRUMANALYZER
            switch (editor.ui.visualizerMode)
            {
                case VISUAL_QUADRASCOPE:
                {
                    scopeData = vs.scopeData;
                    vs.scopeDirty = true;

                    for (i = 0; i < PAULA_CHANNELS; ++i)
                    {
                        scopePointer = editor.muted[i] ? NULL : (scopeData + (i * 40));

                        xPos = 128 + (i * 48);
                        for (x = 0; x < 40; ++x)
                        {
                            yPos = 71 - (scopePointer ? *scopePointer++ : 0);
                            tempScreenBuffer[(yPos * SCREEN_W) + xPos++] = 0x00FFDD00;
                        }
                    }
                }
                break;

                case VISUAL_SPECTRUM:
                {
                    // Draw empty spectrum analyzer dialog
                    for (y = 0; y < SPECTRUM_HEIGHT; ++y)
                    {
                        dstY = ((44 + y) * SCREEN_W) + 120;
                        srcY = y * SPECTRUM_WIDTH;

                        for (x = 0; x < SPECTRUM_WIDTH; ++x)
                        {
                            tempScreenBuffer[dstY + x] = spectrumVisualsBitmap[srcY + x];
                        }
                    }

                    // Draw spectrum bars
                    for (i = 0; i < 23; ++i)
                    {
                        volPeak = editor.spectrumVolumes[i];
                        if (volPeak > 0)
                        {
                            xPos = 129 + (i * 8);

                            for (j = volPeak; j >= 0; --j)
                            {
                                dstY = ((94 - j) * SCREEN_W) + xPos;

                                for (k = 0; k < 6; ++k)
                                {
                                    tempScreenBuffer[dstY + k] = spectrumBarData[j];
                                }
                            }
                        }
                    }
                }
                break;

                default:
                break;
            }

            // DRAW SOME TEXTS
            printThreeDecimals(tempScreenBuffer, mod->currOrder, 72, 3, 0x00000000);
            printThreeDecimals(tempScreenBuffer, mod->head.order[mod->currOrder], 72, 14, 0x00000000);
            printThreeDecimals(tempScreenBuffer, mod->head.orderCount, 72, 25, 0x00000000);
            charOut(tempScreenBuffer, 80, 36, fineTuneChar1[currentSample->fineTune], 0x00000000);
            charOut(tempScreenBuffer, 88, 36, fineTuneChar2[currentSample->fineTune], 0x00000000);
            printTwoHex(tempScreenBuffer, editor.currentSample + 1, 80, 47, 0x00000000);
            printTwoHex(tempScreenBuffer, currentSample->volume, 80, 58, 0x00000000);
            printFiveHex(tempScreenBuffer, currentSample->length, 56, 69, 0x00000000);
            printFiveHex(tempScreenBuffer, currentSample->loopStart, 56, 80, 0x00000000);
            printFiveHex(tempScreenBuffer, currentSample->loopLength, 56, 91, 0x00000000);
        }

        // DRAW SONG NAME
        for (i = 0; i < 20; ++i)
        {
            tempChar = mod->head.moduleTitle[i];
            if (tempChar == '\0')
            {
                charOut(tempScreenBuffer, 104 + (i * FONT_CHAR_W), 102, '_', 0x00000000);
            }
            else
            {
                charOut(tempScreenBuffer, 104 + (i * FONT_CHAR_W), 102, tempChar, 0x00000000);
            }
        }

        // DRAW SONG PLAYING TIME
        printTwoDecimals(tempScreenBuffer, editor.playTime / 3600, 272, 102, 0x00000000);
        printTwoDecimals(tempScreenBuffer, (editor.playTime / 60) % 60, 296, 102, 0x00000000);
   
        if (editor.ui.samplerScreenShown == false)
        {
            // DRAW TEMPO/BPM
            printThreeDecimals(tempScreenBuffer, mod->currBPM, 32, 123, 0x00000000);

            // DRAW CURRENT PATTERN NUMBER
            printTwoDecimals(tempScreenBuffer, mod->currPattern, 8, 127, 0x00000000);
        }

        // DRAW CURRENT SAMPLE'S TEXT
        for (i = 0; i < 22; ++i)
        {
            tempChar = currentSample->text[i];
            if (tempChar == '\0')
            {
                charOut(tempScreenBuffer, 104 + (i * FONT_CHAR_W), 113, '_', 0x00000000);
            }
            else
            {
                charOut(tempScreenBuffer, 104 + (i * FONT_CHAR_W), 113, tempChar, 0x00000000);
            }
        }

        // DRAW MUTE BUTTONS (IF NEEDED)
        if (editor.ui.diskOpScreenShown == false)
        {
            for (i = 0; i < 4; ++i)
            {
                if (editor.muted[i] == true)
                {
                    yPos = 3 + (i * 11);

                    for (y = 0; y < MUTE_HEIGHT; ++y)
                    {
                        dstY = ((yPos + y) * SCREEN_W) + 310;
                        srcY = y * MUTE_WIDTH;

                        for (x = 0; x < MUTE_WIDTH; ++x)
                        {
                            tempScreenBuffer[dstY + x] = muteButtonsBitmap[i][srcY + x];
                        }
                    }
                }
            }
        }

        // DRAW INTRO SCREEN (IF NEEDED)
        if (editor.ui.introScreenShown == true)
        {
            drawIntro(tempScreenBuffer);
        }

        // DRAW DISK OP
        if (editor.ui.diskOpScreenShown == true)
        {
            for (y = 0; y < DISKOP_HEIGHT; ++y)
            {
                dstY = y * SCREEN_W;
                srcY = y * DISKOP_WIDTH;

                for (x = 0; x < DISKOP_WIDTH; ++x)
                {
                    tempScreenBuffer[dstY + x] = diskOpScreenBitmap[srcY + x];
                }
            }

            diskOpDraw(tempScreenBuffer);
        }

        // DRAW SAMPLER SCREEN
        if (editor.ui.samplerScreenShown == true)
        {
            for (y = 0; y < SAMPLER_HEIGHT; ++y)
            {
                dstY = (121 + y) * SCREEN_W;
                srcY = y * SAMPLER_WIDTH;

                for (x = 0; x < SAMPLER_WIDTH; ++x)
                {
                    tempScreenBuffer[dstY + x] = samplerScreenBitmap[srcY + x];
                }
            }

            drawSampler(tempScreenBuffer, currentSample, mod);
        }

        // DRAW CLEAR DIALOG IF NEEDED
        if (editor.ui.clearScreenShown == true)
        {
            for (y = 0; y < CLEAR_HEIGHT; ++y)
            {
                dstY = ((51 + y) * SCREEN_W) + 160;
                srcY = y * CLEAR_WIDTH;

                for (x = 0; x < CLEAR_WIDTH; ++x)
                {
                    tempScreenBuffer[dstY + x] = clearDialogBitmap[srcY + x];
                }
            }
        }

        // DRAW QUIT DIALOG IF NEEDED
        if (editor.ui.exitScreenShown == true)
        {
            for (y = 0; y < QUIT_BOX_HEIGHT; ++y)
            {
                dstY = ((51 + y) * SCREEN_W) + 160;
                srcY = y * QUIT_BOX_WIDTH;

                for (x = 0; x < QUIT_BOX_WIDTH; ++x)
                {
                    tempScreenBuffer[dstY + x] = quitDialogBitmap[srcY + x];
                }
            }
        }

        // DRAW MODULE SIZE
        if (mod->head.moduleSize > 999999)
        {
            printFourDecimals(tempScreenBuffer, mod->head.moduleSize / 1000, 264, 123, 0x00000000);
            charOut(tempScreenBuffer, 296, 123, 'K', 0x00000000);
        }
        else
        {
            printSixDecimals(tempScreenBuffer, mod->head.moduleSize, 264, 123, 0x00000000);
        }

        // DRAW STATUS MESSAGE TEXT
        textOut(tempScreenBuffer, 89, 127, editor.ui.statusMessage, 0x00000000);

        // DRAW MOUSE CURSOR
        for (y = 0; y < POINTER_HEIGHT; ++y)
        {
            if ((input.mouse.y + y) >= SCREEN_H)
            {
                break;
            }

            dstY = (input.mouse.y + y) * SCREEN_W;
            srcY = y * POINTER_WIDTH;

            for (x = 0; x < POINTER_WIDTH; ++x)
            {
                xPos = input.mouse.x + x;
                if (xPos >= SCREEN_W)
                {
                    break;
                }

                paletteIndex = mousePointerBitmap[srcY + x];
                if (paletteIndex != PAL_COLORKEY)
                {
                    tempScreenBuffer[dstY + xPos] = palette[paletteIndex];
                }
            }
        }

        // attempt to sync delta to 60hz
        {
            ticker60HzNow = SDL_GetTicks();
            ticker60HzDelay = ticker60HzNext - ticker60HzNow;

            if (ticker60HzDelay > 0)
            {
                SDL_Delay(ticker60HzDelay);
                ticker60HzNow = SDL_GetTicks();
            }

            ticker60HzWhen = (ticker60HzNow + ticker60HzInterval) / 2;

            ticker60HzNext += ticker60HzInterval;
            if (ticker60HzNext < ticker60HzWhen)
            {
                ticker60HzNext = ticker60HzWhen;
            }
        }

        // COPY TEMP BUFFER TO MAIN FRAMEBUFER
        if (config.blit2x == true)
        {
            frameBufferPointer = frameBuffer + (33 * FULL_SCREEN_W + 40);
            sourceBuffer = tempScreenBuffer;

            srcY = SCREEN_H;
            while (srcY--)
            {
                frameBufferPointerX = frameBufferPointer + FULL_SCREEN_W;

                srcX = SCREEN_W;
                while (srcX--)
                {
                    *frameBufferPointer++ = *sourceBuffer;
                    *frameBufferPointer++ = *sourceBuffer;

                    *frameBufferPointerX++ = *sourceBuffer;
                    *frameBufferPointerX++ = *sourceBuffer++;
                }

                frameBufferPointer += ((FULL_SCREEN_W - SCREEN_W) * 2);
            }

            // Faster than SDL_Flip in software mode
            // Faster than SDL_UpdateRect too (that one does clip+checks)
            SDL_UpdateRects(screen, 1, &tempScreenRect);
        }
        else
        {
            // Faster than SDL_UpperBlit
            SDL_LowerBlit(tempScreenSurface, &frameBufferRect, screen, &frameBufferRect);

            // Faster than SDL_Flip in software mode
            // Faster than SDL_UpdateRect too (that one does clip+checks)
            SDL_UpdateRects(screen, 1, &frameBufferRect);
        }

        frameHandler(mod);
    }

    SDL_WaitThread(thread50Hz, NULL);

    audioClose();
    modFree(mod);

    SDL_FreeSurface(tempScreenSurface);
    SDL_FreeSurface(iconSurface);
    SDL_FreeSurface(screen);

    SDL_Quit();

#ifdef WIN32
    timeEndPeriod(1);
#endif

    return 0;
}

MODULE *createNewMod(void)
{
    int i;

    MODULE *newMod = (MODULE *)calloc(1, sizeof (MODULE));

    for (i = 0; i < 100; ++i)
    {
        newMod->patterns[i] = (modnote_t *)calloc(MOD_ROWS * PAULA_CHANNELS, sizeof (modnote_t));
    }

    newMod->head.patternCount = 1;
    newMod->head.orderCount = 1;
    newMod->MAX_PATTERNS = 100; // M!K!, PT2.x etc

    newMod->head.totalSampleSize = 0;

    newMod->sampleData = (signed char *)calloc(MOD_SAMPLES, 0xFFFF);
    for (i = 0; i < MOD_SAMPLES; ++i)
    {
        newMod->samples[i].offset = 0;
        newMod->samples[i].loopLength = 2;
    }

    calcRealModSize(newMod);

    return newMod;
}
