#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pt_palette.h"
#include "pt_header.h"
#include "pt_sampler.h"
#include "pt_helpers.h"
#include "pt_audio.h"

int modSave(const MODULE *modEntry, const char *fileName)
{
    int i, j, k;
    int sampleDataSize = 0;

    unsigned short tempLoopLength;
    unsigned char tempPatternCount = 0;

    FILE *fmodule;

    pointerSetMode(POINTER_MODE_LOAD, false);

    fmodule = fopen(fileName, "wb");
    if (fmodule == NULL)
    {
        return false;
    }

    fwrite(modEntry->head.moduleTitle, 1, 20, fmodule);

    for (i = 0; i < MOD_SAMPLES; ++i)
    {
        fwrite(modEntry->samples[i].text, 1, 22, fmodule);
        fputc(modEntry->samples[i].length >> 9, fmodule);
        fputc(modEntry->samples[i].length >> 1, fmodule);
        fputc(modEntry->samples[i].fineTune & 0x0F, fmodule);
        fputc(modEntry->samples[i].volume, fmodule);
        fputc(modEntry->samples[i].loopStart >> 9, fmodule);
        fputc(modEntry->samples[i].loopStart >> 1, fmodule);

        tempLoopLength = modEntry->samples[i].loopLength;
        if (tempLoopLength < 2)
        {
            tempLoopLength = 2;
        }

        fputc(tempLoopLength >> 9, fmodule);
        fputc(tempLoopLength >> 1, fmodule);

        sampleDataSize += (modEntry->samples[i].length > 2) ? modEntry->samples[i].length : 0;
    }

    fputc(modEntry->head.orderCount, fmodule);
    fputc(0x7F, fmodule); // ProTracker puts 0x7F at this place (restart pos)
    fwrite(modEntry->head.order, 1, MOD_ORDERS, fmodule);

    tempPatternCount = 0;
    for (i = 0; i < MOD_ORDERS; ++i)
    {
        if (tempPatternCount < modEntry->head.order[i])
        {
            tempPatternCount = modEntry->head.order[i];
        }
    }

    if (tempPatternCount > 100)
    {
        tempPatternCount = 100;
    }

    fwrite((tempPatternCount <= 64) ? "M.K." : "M!K!", 1, 4, fmodule);

    for (i = 0; i < (tempPatternCount + 1); ++i)
    {
        for (j = 0; j < MOD_ROWS; ++j)
        {
            for (k = 0; k < PAULA_CHANNELS; ++k)
            {
                modnote_t tmp = modEntry->patterns[i][(j * PAULA_CHANNELS) + k];

                fputc(((tmp.period >> 8) & 0x0F) | (tmp.sample & 0x10), fmodule);
                fputc(tmp.period & 0xFF, fmodule);
                fputc(((tmp.sample & 0x0F) << 4) | (tmp.command & 0x0F), fmodule);
                fputc(tmp.param, fmodule);
            }
        }
    }

    fwrite(modEntry->sampleData, 1, sampleDataSize, fmodule);

    editor.currentSample = 0;

    pointerSetMode(POINTER_MODE_IDLE, true);

    return true;
}

static int checkModType(const char *buf)
{
    if (!strncmp(buf, "M.K.", 4))
    {
        return FORMAT_MK; // ProTracker v1.x
    }
    else if (!strncmp(buf, "M!K!", 4))
    {
        return FORMAT_MK2; // ProTracker v2.x (if >64 patterns)
    }
    else if (!strncmp(buf, "FLT4", 4)) 
    {
        return FORMAT_FLT4; // StarTrekker (4 channel MODs only)
    }
    else if (!strncmp(buf, "4CHN", 4))
    {
        return FORMAT_4CHN; // FastTracker II (4 channel MODs only)
    }
    else if (!strncmp(buf, "N.T.", 4))
    {
        return FORMAT_MK; // NoiseTracker 1.0, same as ProTracker v1.x (?)
    }

    return FORMAT_UNKNOWN; // May be The Ultimate SoundTracker, 15 samples
}

MODULE *modLoad(const char *fileName)
{
    int i, j, pattern, row, channel;
    int sampleOffset = 0;
    int mightBeSTK = false;
    int numSamples;
    unsigned char bytes[4];

    char modSig[4];

    FILE *fmodule = NULL;
    MODULE *newModule = NULL;
    modnote_t *note = NULL;

    pointerSetMode(POINTER_MODE_LOAD, false);

    newModule = (MODULE *)calloc(1, sizeof (MODULE));
    if (newModule == NULL)
    {
        return NULL; // out of memory
    }

    fmodule = fopen(fileName, "rb");
    if (fmodule == NULL)
    {
        free(newModule);

        return NULL; // can't be opened
    }

    fseek(fmodule, 0, SEEK_END);
    newModule->head.moduleSize = ftell(fmodule);
    fseek(fmodule, 0, SEEK_SET);

    if (newModule->head.moduleSize <= 1624)
    {
        free(newModule);

        return NULL; // definitely not a MOD
    }

    // check module tag
    fseek(fmodule, 0x0438, SEEK_SET);
    fread(modSig, 1, 4, fmodule);

    newModule->head.format = checkModType(modSig);
    switch (newModule->head.format)
    {
        case FORMAT_MK: newModule->MAX_PATTERNS = 64; break;
        case FORMAT_MK2: newModule->MAX_PATTERNS = 100; break;
        case FORMAT_FLT4: newModule->MAX_PATTERNS = 100; break;
        case FORMAT_4CHN: newModule->MAX_PATTERNS = 100; break;

        case FORMAT_UNKNOWN:
            newModule->MAX_PATTERNS = 64;
            mightBeSTK = true;
        break;
    }

    fseek(fmodule, 0, SEEK_SET);

    fread(newModule->head.moduleTitle, 1, 20, fmodule);
    for (i = 0; i < 20; ++i)
    {
        char ch = newModule->head.moduleTitle[i];
        if (((ch < ' ') || (ch > '~')) && (ch != '\0'))
        {
            newModule->head.moduleTitle[i] = ' ';
        }
    }

    // read sample information
    for (i = 0; i < MOD_SAMPLES; ++i)
    {
        if ((mightBeSTK == true) && (i > 14))
        {
            newModule->samples[i].loopLength = 2;
        }
        else
        {
            fread(newModule->samples[i].text, 1, 22, fmodule);
            for (j = 0; j < 22; ++j)
            {
                char ch = newModule->samples[i].text[j];
                if (((ch < ' ') || (ch > '~')) && (ch != '\0'))
                {
                    newModule->samples[i].text[j] = ' ';
                }
            }

            newModule->samples[i].length = fGetWordBigEndian(fmodule) * 2;

            fread(&newModule->samples[i].fineTune, 1, 1, fmodule);
            newModule->samples[i].fineTune = newModule->samples[i].fineTune & 0x0F;

            fread(&newModule->samples[i].volume, 1, 1, fmodule);
            if (newModule->samples[i].volume > 64)
            {
                newModule->samples[i].volume = 64;
            }

            if (mightBeSTK == true)
            {
                // In .STK files, "loop start" is in bytes, not samples
                newModule->samples[i].loopStart = fGetWordBigEndian(fmodule);
            }
            else
            {
                // In .MOD files, "loop start" is in samples (word), not bytes
                newModule->samples[i].loopStart = fGetWordBigEndian(fmodule) * 2;
            }

            // The difference above does not apply to "loop length"/"sample length"
            newModule->samples[i].loopLength = fGetWordBigEndian(fmodule) * 2;

            if (newModule->samples[i].loopLength < 2)
            {
                newModule->samples[i].loopLength = 2;
            }

            newModule->head.totalSampleSize += newModule->samples[i].length;
        }
    }

    fread(&newModule->head.orderCount, 1, 1, fmodule);
    if ((newModule->head.orderCount == 0) || (newModule->head.orderCount > 128))
    {
        free(newModule);
        fclose(fmodule);

        return NULL;
    }

    newModule->head.restartPos = fgetc(fmodule);
    if ((mightBeSTK == true) && (newModule->head.restartPos != 120))
    {
        // That's it, this is not a proper MOD
        // In very rare occasion this byte could still be 120
        // while the file is not a MOD.
        // But that will be really rare, so let's not think
        // about that.

        free(newModule);
        fclose(fmodule);

        return NULL;
    }

    if (mightBeSTK == true)
    {
        // If we're still here at this point and
        // the mightBeSTK flag is set,
        // then it's definitely a proper
        // The Ultimate SoundTracker module.

        newModule->head.format = FORMAT_STK;
    }

    for (i = 0; i < MOD_ORDERS; ++i)
    {
        fread(&newModule->head.order[i], 1, 1, fmodule);

        if (newModule->head.order[i] > newModule->head.patternCount)
        {
            newModule->head.patternCount = newModule->head.order[i];
        }
    }

    newModule->head.patternCount++;

    if (newModule->head.format != FORMAT_STK) // The Ultimate SoundTracker MODs doesn't have this tag
    {
        fseek(fmodule, 4, SEEK_CUR); // We already read/tested the tag earlier, skip it
    }

    // init MAX_PATTERN of patterns and load patternCount of patterns
    for (pattern = 0; pattern < newModule->MAX_PATTERNS; ++pattern)
    {
        newModule->patterns[pattern] = (modnote_t *)calloc(MOD_ROWS * PAULA_CHANNELS, sizeof (modnote_t));
        if (newModule->patterns[pattern] == NULL)
        {
            fclose(fmodule);
            free(newModule);

            return NULL; // out of memory
        }
    }

    for (pattern = 0; pattern < newModule->head.patternCount; ++pattern)
    {
        note = newModule->patterns[pattern];
        for (row = 0; row < MOD_ROWS; ++row)
        {
            for (channel = 0; channel < PAULA_CHANNELS; ++channel)
            {
                fread(bytes, 1, 4, fmodule);

                note->period = (LO_NYBBLE(bytes[0]) << 8) | bytes[1];
                note->sample = (bytes[0] & 0xF0) | HI_NYBBLE(bytes[2]); // Don't (!) clamp, the player checks for invalid sample data
                note->command = LO_NYBBLE(bytes[2]);
                note->param = bytes[3];

                if (newModule->head.format == FORMAT_4CHN)
                {
                    // Remove FastTracker II 8xx/E8x panning commands if present
                    if (note->command == 0x08)
                    {
                        // 8xx
                        note->command = 0;
                        note->param = 0;
                    }
                    else if ((note->command == 0x0E) && (HI_NYBBLE(note->param) == 0x08))
                    {
                        // E8x
                        note->command = 0;
                        note->param = 0;
                    }

                    // Remove F00, FastTracker II didn't use F00 as STOP
                    if ((note->command == 0x0F) && (note->param == 0x00))
                    {
                        note->command = 0;
                        note->param = 0;
                    }
                }

                note++;
            }
        }
    }

    newModule->sampleData = (signed char *)malloc(newModule->head.totalSampleSize);
    if (newModule->sampleData == NULL)
    {
        fclose(fmodule);
        free(newModule);

        return NULL; // out of memory
    }

    // load sample data
    numSamples = (newModule->head.format == FORMAT_STK) ? 15 : 31;
    for (i = 0; i < numSamples; ++i)
    {
        newModule->samples[i].offset = sampleOffset;
        fread(&newModule->sampleData[sampleOffset], 1, newModule->samples[i].length, fmodule);
        sampleOffset += newModule->samples[i].length;
    }

    fclose(fmodule);

    // calculate MOD size
    calcRealModSize(newModule);

    // remember to unmute all channels
    editor.muted[0] = false;
    editor.muted[1] = false;
    editor.muted[2] = false;
    editor.muted[3] = false;

    editor.useLEDFilter = false; // the effect E0x in pattern data decides this, pre-null it!

    editor.currentSample = 0;
    editor.playTime = 0;

    setupSampler(newModule);

    pointerSetMode(POINTER_MODE_IDLE, true);

    return newModule;
}
