#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "pt_header.h"
#include "pt_audio.h"
#include "pt_helpers.h"
#include "pt_tables.h"
#include "pt_modloader.h"
#include "pt_config.h"

static unsigned char modSpeed, modBPM;
static unsigned char modTick, modPattern, modOrder;
static signed char modRow; // must be signed

static signed int sampleCounter = 0;
static unsigned int samplesPerTick;

static unsigned char PBreakPosition = 0;
static unsigned char PattDelayTime = 0, PattDelayTime2 = 0;
static int PosJumpAssert = false, PBreakFlag = false;

static signed short tempPeriod;
static signed char tempVolume;
static unsigned char tempFlags;

static MODULE *source = NULL;

inline static int periodTonote(int fineTune, int period)
{
    int i;
    for (i = 0; i < 36; ++i)
    {
        if (amigaPeriods[fineTune][i] <= period)
        {
            break;
        }
    }

    return i;
}

enum
{
    FLAG_NOTE = 1,
    FLAG_SAMPLE = 2,
    FLAG_NEWSAMPLE = 4,
    FLAG_SET_VUMETER = 8
};

enum
{
    TEMPFLAG_START = 1,
    TEMPFLAG_DELAY = 2,
    TEMPFLAG_SET_VUMETER = 4,
    TEMPFLAG_NEW_SAMPLE = 8
};


// Normal effects
static void fxArpeggio(mod_channel *ch);
static void fxPortamentoSlideUp(mod_channel *ch);
static void fxPortamentoSlideDown(mod_channel *ch);
static void fxGlissando(mod_channel *ch);
static void fxVibrato(mod_channel *ch);
static void fxGlissandoVolumeSlide(mod_channel *ch);
static void fxVibratoVolumeSlide(mod_channel *ch);
static void fxTremolo(mod_channel *ch);
static void fxNotInUse(mod_channel *ch);
static void fxSampleOffset(mod_channel *ch);
static void fxVolumeSlide(mod_channel *ch);
static void fxPositionJump(mod_channel *ch);
static void fxSetVolume(mod_channel *ch);
static void fxPatternBreak(mod_channel *ch);
static void fxExtended(mod_channel *ch);
static void fxSetTempo(mod_channel *ch);

// Extended effects
static void efxSetLEDFilter(mod_channel *ch);
static void efxFinePortamentoSlideUp(mod_channel *ch);
static void efxFinePortamentoSlideDown(mod_channel *ch);
static void efxGlissandoControl(mod_channel *ch);
static void efxVibratoControl(mod_channel *ch);
static void efxSetFineTune(mod_channel *ch);
static void efxPatternLoop(mod_channel *ch);
static void efxTremoloControl(mod_channel *ch);
static void efxKarplusStrong(mod_channel *ch);
static void efxRetrigNote(mod_channel *ch);
static void efxFineVolumeSlideUp(mod_channel *ch);
static void efxFineVolumeSlideDown(mod_channel *ch);
static void efxNoteCut(mod_channel *ch);
static void efxNoteDelay(mod_channel *ch);
static void efxPatternDelay(mod_channel *ch);
static void efxInvertLoop(mod_channel *ch);

typedef void (*effect_routine)(mod_channel *);

// Normal effects
static effect_routine fxRoutines[16] =
{
    fxArpeggio,
    fxPortamentoSlideUp,
    fxPortamentoSlideDown,
    fxGlissando,
    fxVibrato,
    fxGlissandoVolumeSlide,
    fxVibratoVolumeSlide,
    fxTremolo,
    fxNotInUse,
    fxSampleOffset,
    fxVolumeSlide,
    fxPositionJump,
    fxSetVolume,
    fxPatternBreak,
    fxExtended,
    fxSetTempo
};

// Extended effects
static effect_routine efxRoutines[16] =
{
    efxSetLEDFilter,
    efxFinePortamentoSlideUp,
    efxFinePortamentoSlideDown,
    efxGlissandoControl,
    efxVibratoControl,
    efxSetFineTune,
    efxPatternLoop,
    efxTremoloControl,
    efxKarplusStrong,
    efxRetrigNote,
    efxFineVolumeSlideUp,
    efxFineVolumeSlideDown,
    efxNoteCut,
    efxNoteDelay,
    efxPatternDelay,
    efxInvertLoop
};

void setPlayerModSource(MODULE *modEntry)
{
    source = modEntry;
}

static void processInvertLoop(mod_channel *ch)
{
    if (ch->invertLoopSpeed > 0)
    {
        ch->invertLoopDelay += invertLoopSpeeds[ch->invertLoopSpeed];
        if (ch->invertLoopDelay >= 128)
        {
            ch->invertLoopDelay = 0;

            if (ch->sample != 0)
            {
                signed char invertLoopTemp;
                signed char *invertLoopData;

                MODULE_SAMPLE *s = &source->samples[ch->sample - 1];

                if (s->loopLength >= 4)
                {
                    ch->invertLoopOffset++;
                    if (ch->invertLoopOffset >= (s->loopStart + s->loopLength))
                    {
                        ch->invertLoopOffset = s->loopStart;
                    }

                    invertLoopData = &source->sampleData[s->offset + ch->invertLoopOffset];
                    invertLoopTemp = -1 - *invertLoopData;
                    *invertLoopData = invertLoopTemp;
                }
            }
        }
    }
}

static void efxSetLEDFilter(mod_channel *ch)
{
    if (modTick == 0)
    {
        editor.useLEDFilter = !(ch->param & 1);
    }
}

static void efxFinePortamentoSlideUp(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (tempPeriod > 0)
        {
            ch->period -= LO_NYBBLE(ch->param);

            if (ch->period < PT_MIN_PERIOD)
            {
                ch->period = PT_MIN_PERIOD;
            }

            tempPeriod = ch->period;
        }
    }
}

static void efxFinePortamentoSlideDown(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (tempPeriod > 0)
        {
            ch->period += LO_NYBBLE(ch->param);

            if (ch->period > PT_MAX_PERIOD)
            {
                ch->period = PT_MAX_PERIOD;
            }

            tempPeriod = ch->period;
        }
    }
}

static void efxGlissandoControl(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->glissandoControl = LO_NYBBLE(ch->param);
    }
}

static void efxVibratoControl(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->vibratoControl = LO_NYBBLE(ch->param);
    }
}

static void efxSetFineTune(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->fineTune = LO_NYBBLE(ch->param);
    }
}

static void efxPatternLoop(mod_channel *ch)
{
    if (modTick == 0)
    {
        unsigned char tempParam = LO_NYBBLE(ch->param);
        if (tempParam == 0)
        {
            ch->patternLoopRow = modRow;

            return;
        }

        if (ch->patternLoopCounter == 0)
        {
            ch->patternLoopCounter = tempParam;
        }
        else
        {
            ch->patternLoopCounter--;
            if (ch->patternLoopCounter == 0)
            {
                return;
            }
        }

        PBreakPosition = ch->patternLoopRow;
        PBreakFlag = true;
    }
}

static void efxTremoloControl(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->tremoloControl = LO_NYBBLE(ch->param);
    }
}

static void efxKarplusStrong(mod_channel *ch)
{
    if (ch->sample > 0)
    {
        MODULE_SAMPLE *s = &source->samples[ch->sample - 1];

        signed char *sampleLoopData = source->sampleData + s->offset + s->loopStart;

        unsigned int loopLength = s->loopLength - 2;
        unsigned int loopLengthCounter = loopLength;

        while (loopLengthCounter--)
        {
            *sampleLoopData = (*sampleLoopData + *(sampleLoopData + 1)) / 2;
            sampleLoopData++;
        }

        *sampleLoopData = (*sampleLoopData + *(sampleLoopData - loopLength)) / 2;
    }
}

static void efxRetrigNote(mod_channel *ch)
{
    unsigned char retrigTick = LO_NYBBLE(ch->param);

    if (retrigTick > 0)
    {
        if ((modTick % retrigTick) == 0)
        {
            tempFlags |= TEMPFLAG_START;
        }
    }
}

static void efxFineVolumeSlideUp(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->volume += LO_NYBBLE(ch->param);

        if (ch->volume > 64)
        {
            ch->volume = 64;
        }

        tempVolume = ch->volume;
    }
}

static void efxFineVolumeSlideDown(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->volume -= LO_NYBBLE(ch->param);

        if (ch->volume < 0)
        {
            ch->volume = 0;
        }

        tempVolume = ch->volume;
    }
}

static void efxNoteCut(mod_channel *ch)
{
    if (modTick == LO_NYBBLE(ch->param))
    {
        ch->volume = 0;
        tempVolume = 0;
    }
}

static void efxNoteDelay(mod_channel *ch)
{
    unsigned char delayTick = LO_NYBBLE(ch->param);

    if (modTick == 0)
    {
        ch->tempFlagsBackup = tempFlags;
    }

    if (modTick < delayTick)
    {
        tempFlags = TEMPFLAG_DELAY;
    }
    else if (modTick == delayTick)
    {
        tempFlags = ch->tempFlagsBackup;
    }
}

static void efxPatternDelay(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (PattDelayTime2 == 0)
        {
            PattDelayTime = LO_NYBBLE(ch->param) + 1;
        }
    }
}

static void efxInvertLoop(mod_channel *ch)
{
    if (modTick == 0)
    {
        ch->invertLoopSpeed = LO_NYBBLE(ch->param);
        processInvertLoop(ch);
    }
}

static void handleGlissando(mod_channel *ch)
{
    if (tempPeriod > 0)
    {
        if (ch->period < ch->tempPeriod)
        {
            ch->period += ch->glissandoSpeed;

            if (ch->period > ch->tempPeriod)
            {
                ch->period = ch->tempPeriod;
            }
        }
        else
        {
            ch->period -= ch->glissandoSpeed;

            if (ch->period < ch->tempPeriod)
            {
                ch->period = ch->tempPeriod;
            }
        }

        if (ch->glissandoControl != 0)
        {
            // Note-based slide, less smooth
            int i;
            short *tablePointer = (short *)&rawAmigaPeriods[ch->fineTune * 37];
            for (i = 0; i < 36; ++i)
            {
                if (tablePointer[i] <= ch->period)
                {
                    tempPeriod = tablePointer[i];
                    return;
                }
            }
        }
        else
        {
            // Normal portamento slide
            tempPeriod = ch->period;
        }
    }
}

static void processVibrato(mod_channel *ch)
{
    if (tempPeriod > 0)
    {
        int vibratoData;
        unsigned char vibratoTemp;

        vibratoTemp = ch->vibratoPos / 4; // 0..64
        vibratoTemp &= 0x1F; // wrap 0..32, two rounds - one for neg, one for pos

        switch (ch->vibratoControl & 3)
        {
            case 0: // 0 - sine
            {
                vibratoData = sinusTable[vibratoTemp];
            }
            break;

            case 1: // 1 - ramp down
            {
                if (ch->vibratoPos < 128)
                {
                    vibratoData = vibratoTemp * 8;
                }
                else
                {
                    vibratoData = 255 - (vibratoTemp * 8);
                }
            }
            break;

            default: // 2 - square wave
            {
                vibratoData = 255;
            }
            break;
        }

        vibratoData = (vibratoData * ch->vibratoDepth) / 128;

        if (ch->vibratoPos < 128)
        {
            tempPeriod += (short)vibratoData;
            if (tempPeriod > 907)
            {
                tempPeriod = 907;
            }
        }
        else
        {
            tempPeriod -= (short)vibratoData;
            if (tempPeriod < 108)
            {
                tempPeriod = 108;
            }
        }

        ch->vibratoPos += (ch->vibratoSpeed * 4);
        // vibratoPos is 8-bit and wraps around 255
    }
}

static void processTremolo(mod_channel *ch)
{
    if (tempVolume > 0)
    {
        int tremoloData;
        unsigned char tremoloTemp;
        
        tremoloTemp = ch->tremoloPos / 4; // 0..64
        tremoloTemp &= 0x1F; // wrap 0..32, two rounds - one for neg, one for pos

        switch (ch->tremoloControl & 3)
        {
            case 0: // 0 - sine
            {
                tremoloData = sinusTable[tremoloTemp];
            }
            break;

            case 1: // 1 - ramp down
            {
                // It's supposed to use the tremolo pos here,
                // but instead PT uses the vibrato pos...
                // How stupid, but I will of course implement the qurik,
                // because it could make a major sound difference.

                if (ch->vibratoPos < 128)
                {
                    tremoloData = tremoloTemp * 8;
                }
                else
                {
                    tremoloData = 255 - (tremoloTemp * 8);
                }
            }
            break;

            default: // 2 - square wave
            {
                tremoloData = 255;
            }
            break;
        }

        tremoloData = (tremoloData * ch->tremoloDepth) / 64;

        if (ch->tremoloPos < 128)
        {
            tempVolume += (char)tremoloData;
            if (tempVolume > 64)
            {
                tempVolume = 64;
            }
        }
        else
        {
            tempVolume -= (char)tremoloData;
            if (tempVolume < 0)
            {
                tempVolume = 0;
            }
        }

        ch->tremoloPos += (ch->tremoloSpeed * 4);
        // tremoloPos is 8-bit and wraps around 255
    }
}

static void fxArpeggio(mod_channel *ch)
{
    int i;
    int noteToAdd;
    short *tablePointer;

    int arpeggioTick = modTick % 3;
    if (arpeggioTick == 0)
    {
        tempPeriod = ch->period;
        return;
    }
    else if (arpeggioTick == 1)
    {
        noteToAdd = HI_NYBBLE(ch->param);
    }
    else if (arpeggioTick == 2)
    {
        noteToAdd = LO_NYBBLE(ch->param);
    }

    tablePointer = (short *)&rawAmigaPeriods[ch->fineTune * 37];
    for (i = 0; i < 36; ++i)
    {
        if (tablePointer[i] <= ch->period)
        {
            tempPeriod = tablePointer[i + noteToAdd];
            return;
        }
    }
}

static void fxPortamentoSlideUp(mod_channel *ch)
{
    if ((modTick > 0) && (tempPeriod > 0))
    {
        ch->period -= ch->param;

        if (ch->period < PT_MIN_PERIOD)
        {
            ch->period = PT_MIN_PERIOD;
        }

        tempPeriod = ch->period;
    }
}

static void fxPortamentoSlideDown(mod_channel *ch)
{
    if ((modTick > 0) && (tempPeriod > 0))
    {
        ch->period += ch->param;

        if (ch->period > PT_MAX_PERIOD)
        {
            ch->period = PT_MAX_PERIOD;
        }

        tempPeriod = ch->period;
    }
}

static void fxGlissando(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (ch->param != 0)
        {
            ch->glissandoSpeed = ch->param;
        }
    }
    else
    {
        handleGlissando(ch);
    }
}

static void fxVibrato(mod_channel *ch)
{
    if (modTick == 0)
    {
        unsigned char hiNybble = HI_NYBBLE(ch->param);
        unsigned char loNybble = LO_NYBBLE(ch->param);

        if (hiNybble != 0)
        {
            ch->vibratoSpeed = hiNybble;
        }

        if (loNybble != 0)
        {
            ch->vibratoDepth = loNybble;
        }
    }
    else
    {
        processVibrato(ch);
    }
}

static void fxGlissandoVolumeSlide(mod_channel *ch)
{
    if (modTick > 0)
    {
        handleGlissando(ch);
        fxVolumeSlide(ch);
    }
}

static void fxVibratoVolumeSlide(mod_channel *ch)
{
    if (modTick > 0)
    {
        processVibrato(ch);
        fxVolumeSlide(ch);
    }
}

static void fxTremolo(mod_channel *ch)
{
    if (modTick == 0)
    {
        unsigned char hiNybble = HI_NYBBLE(ch->param);
        unsigned char loNybble = LO_NYBBLE(ch->param);

        if (hiNybble > 0)
        {
            ch->tremoloSpeed = hiNybble;
        }

        if (loNybble > 0)
        {
            ch->tremoloDepth = loNybble;
        }
    }
    else
    {
        processTremolo(ch);
    }
}

static void fxNotInUse(mod_channel *ch)
{
    // This is an empty effect slot not used in
    // protracker... often used for demo efx syncing
    //
    // Used for panning in fasttracker and later PC
    // MOD players/trackers
}

static void fxSampleOffset(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (ch->param > 0)
        {
            ch->offsetTemp = ch->param * 256;
        }

        ch->offset += ch->offsetTemp;
    }
}

static void fxVolumeSlide(mod_channel *ch)
{
    if (modTick > 0)
    {
        unsigned char hiNybble = HI_NYBBLE(ch->param);
        unsigned char loNybble = LO_NYBBLE(ch->param);

        if (hiNybble == 0)
        {
            ch->volume -= loNybble;
            if (ch->volume < 0)
            {
                ch->volume = 0;
            }

            tempVolume = ch->volume;
        }
        else
        {
            ch->volume += hiNybble;
            if (ch->volume > 64)
            {
                ch->volume = 64;
            }

            tempVolume = ch->volume;
        }
    }
}

static void fxPositionJump(mod_channel *ch)
{
    modOrder = ch->param - 1;
    PBreakPosition = 0;
    PosJumpAssert = true;
}

static void fxSetVolume(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (ch->param > 64)
        {
            ch->param = 64;
        }

        ch->volume = ch->param;
        tempVolume = ch->param;
    }
}

static void fxPatternBreak(mod_channel *ch)
{
    unsigned char pos = ((HI_NYBBLE(ch->param) * 10) + LO_NYBBLE(ch->param));

    if (pos > 63)
    {
        pos = 0;
    }

    PBreakPosition = pos;
    PosJumpAssert = true;
}

static void fxExtended(mod_channel *ch)
{
    efxRoutines[HI_NYBBLE(ch->param)](ch);
}

static void fxSetTempo(mod_channel *ch)
{
    if (modTick == 0)
    {
        if (ch->param < 32)
        {
            if (ch->param == 0)
            {
                modStop();
            }
            else
            {
                modSetSpeed(ch->param);
            }
        }
        else
        {
            modSetTempo(ch->param);
        }
    }
}

static void processEffects(mod_channel *ch)
{
    if (modTick > 0)
    {
        processInvertLoop(ch);
    }

    if (!(!ch->command && !ch->param))
    {
        fxRoutines[ch->command](ch);
    }
}

static void fetchPatternData(mod_channel *ch)
{
    modnote_t *note = &source->patterns[modPattern][(modRow * PAULA_CHANNELS) + ch->chanIndex];

    if (note->sample > 0)
    {
        if (ch->sample != note->sample)
        {
            ch->flags |= FLAG_NEWSAMPLE;
        }

        ch->sample = note->sample;
        ch->flags |= FLAG_SAMPLE;
        ch->fineTune = source->samples[ch->sample - 1].fineTune;
    }

    ch->command = note->command;
    ch->param = note->param;

    if (note->period > 0)
    {
        int tempNote;

        // set finetune if E5x is called
        if (ch->command == 0xE)
        {
            if (HI_NYBBLE(ch->param) == 0x5)
            {
                ch->fineTune = LO_NYBBLE(ch->param);
            }
        }

        tempNote = periodTonote(0, CLAMP(note->period, 113, 856));

        ch->noNote = false;

        ch->tempPeriod = amigaPeriods[ch->fineTune][tempNote];
        ch->flags |= FLAG_NOTE;
    }
    else
    {
        ch->noNote = true;
    }
}

static void processChannel(mod_channel *ch)
{
    tempFlags = 0;

    if (modTick == 0)
    {
        if (PattDelayTime2 == 0)
        {
            fetchPatternData(ch);
        }

        // In PT, the spectrum analyzer doesn't use the last volume affected
        // by volume effects. It just uses the last s->sample value for
        // the current channel.
        // So we use a separate variable for this, and update it before
        // we update the spectrum analyzer for correct visual results.
        if (ch->flags & FLAG_SAMPLE)
        {
            if (ch->sample > 0)
            {
                ch->rawVolume = source->samples[ch->sample - 1].volume;
            }
        }

        if (ch->flags & FLAG_NOTE)
        {
            ch->flags &= ~FLAG_NOTE;

            if ((ch->command != 0x03) && (ch->command != 0x05))
            {
                ch->period = ch->tempPeriod;

                if (ch->sample > 0)
                {
                    tempFlags |= TEMPFLAG_START;
                }

                // Update spectrum bars if neeeded
                if ((editor.ui.visualizerMode == VISUAL_SPECTRUM) && (editor.muted[ch->chanIndex] == false))
                {
                    unsigned int scaledVol, scaledVol2;
                    unsigned int scaledNote;

                    scaledVol = ch->rawVolume;
                    scaledVol = (scaledVol * 256) / 682;
                    scaledVol2 = scaledVol / 2;

                    scaledNote = 743 - (ch->period - 113);
                    scaledNote = (scaledNote * scaledNote) / 25093;
                    if (scaledNote > 22)
                    {
                        scaledNote = 22;
                    }

                    editor.spectrumVolumes[scaledNote] += scaledVol;
                    if (editor.spectrumVolumes[scaledNote] > 35)
                    {
                        editor.spectrumVolumes[scaledNote] = 35;
                    }

                    if (scaledNote > 0)
                    {
                        int lookUp = scaledNote - 1;

                        editor.spectrumVolumes[lookUp] += scaledVol2;
                        if (editor.spectrumVolumes[lookUp] > 35)
                        {
                            editor.spectrumVolumes[lookUp] = 35;
                        }
                    }

                    if (scaledNote < 22)
                    {
                        int lookUp = scaledNote + 1;

                        editor.spectrumVolumes[lookUp] += scaledVol2;
                        if (editor.spectrumVolumes[lookUp] > 35)
                        {
                            editor.spectrumVolumes[lookUp] = 35;
                        }
                    }
                }
            }

            tempFlags |= TEMPFLAG_SET_VUMETER; 

            ch->tempFlagsBackup = 0;

            if ((ch->vibratoControl & 4) == 0)
            {
                ch->vibratoPos = 0;
            }

            if ((ch->tremoloControl & 4) == 0)
            {
                ch->tremoloPos = 0;
            }
        }

        if (ch->flags & FLAG_SAMPLE)
        {
            ch->flags &= ~FLAG_SAMPLE;

            if (ch->sample > 0)
            {
                MODULE_SAMPLE *s = &source->samples[ch->sample - 1];

                ch->volume = s->volume;
                ch->invertLoopOffset = s->loopStart;

                if ((ch->command != 0x03) && (ch->command != 0x05))
                {
                    ch->offset = 0;
                    ch->offsetBugNotAdded = false;
                }

                if (ch->flags & FLAG_NEWSAMPLE)
                {
                    ch->flags &= ~FLAG_NEWSAMPLE;

                    if ((ch->period > 0) && ((ch->noNote == true)
                        || ch->command == 0x03
                        || ch->command == 0x05))
                    {
                        tempFlags |= TEMPFLAG_NEW_SAMPLE;
                    }
                }
            }
        }
    }

    if (ch->flags & FLAG_SET_VUMETER)
    {
        ch->flags &= ~FLAG_SET_VUMETER;

        tempFlags |= TEMPFLAG_SET_VUMETER; 
    }

    tempPeriod = ch->period;
    tempVolume = ch->volume;

    processEffects(ch);

    if (!(tempFlags & TEMPFLAG_DELAY))
    {
        if (tempFlags & TEMPFLAG_SET_VUMETER)
        {
            if (editor.muted[ch->chanIndex] == false)
            {
                editor.vuMeterVolumes[ch->chanIndex] = vuMeterHeights[tempVolume];
            }
        }

        if (tempFlags & TEMPFLAG_NEW_SAMPLE)
        {
            if (ch->sample > 0)
            {
                MODULE_SAMPLE *s = &source->samples[ch->sample - 1];
                if (s->length > 2)
                {
                    mixerSwapChSource(ch->chanIndex, &source->sampleData[s->offset], s->length, s->loopStart, s->loopLength);
                }
                else
                {
                    mixerSetChSource(ch->chanIndex, NULL, 0, 0, 0, 0);
                }
            }
        }
        else if (tempFlags & TEMPFLAG_START)
        {
            if (ch->sample > 0)
            {
                MODULE_SAMPLE *s = &source->samples[ch->sample - 1];

                if (s->length > 2)
                {
                    if (ch->offset > 0)
                    {
                        mixerSetChSource(ch->chanIndex, source->sampleData + s->offset, s->length, s->loopStart, s->loopLength, ch->offset);

                        if (ch->offsetBugNotAdded == false)
                        {
                            ch->offset += ch->offsetTemp; // emulates add sample offset bug (fixes no_mercy.mod)
                            ch->offsetBugNotAdded = true;
                        }
                    }
                    else
                    {
                        mixerSetChSource(ch->chanIndex, source->sampleData + s->offset, s->length, s->loopStart, s->loopLength, 0);
                    }
                }
                else
                {
                    mixerSetChSource(ch->chanIndex, NULL, 0, 0, 0, 0);
                }
            }
        }

        mixerSetChVol(ch->chanIndex, (float)tempVolume);

        if (tempPeriod != 0)
        {
            if ((tempPeriod >= 108) && (tempPeriod <= 907)) // should never happen, but just in case
            {
                mixerSetChRate(ch->chanIndex, frequencyTable[(int)tempPeriod]);
            }
        }
        else
        {
            mixerSetChVol(ch->chanIndex, 0.0f); // arp override bugfix
        }
    }
}

static void nextPosition(void)
{
    modRow = PBreakPosition;

    PBreakPosition = 0;
    PosJumpAssert = false;

    if (editor.playMode == PLAY_MODE_NORMAL)
    {
        int pos = modOrder + 1;

        if (pos >= source->head.orderCount)
        {
            pos = 0;
        }

        modOrder = pos;

        modPattern = source->head.order[modOrder];
        if (modPattern >= source->MAX_PATTERNS)
        {
            modPattern = source->MAX_PATTERNS - 1;
        }
    }
    else
    {
        // fix for pattern play mode
        modOrder = source->currOrder;
    }
}

static void processTick(void)
{
    processChannel(source->channels);
    processChannel(source->channels + 1);
    processChannel(source->channels + 2);
    processChannel(source->channels + 3);

    if (editor.songPlaying == true)
    {
        modTick++;
        if (modTick >= modSpeed)
        {
            modTick = 0;

            source->currRow = modRow;

            modRow++;

            if (PattDelayTime > 0)
            {
                PattDelayTime2 = PattDelayTime;
                PattDelayTime = 0;
            }

            if (PattDelayTime2 > 0)
            {
                PattDelayTime2--;
                if (PattDelayTime2 != 0)
                {
                    modRow--;
                }
            }

            if (PBreakFlag == true)
            {
                PBreakFlag = false;

                modRow = PBreakPosition;
                PBreakPosition = 0;
            }

            if ((modRow >= MOD_ROWS) || (PosJumpAssert == true))
            {
                nextPosition();
            }

            source->currRow = modRow;
            source->currOrder = modOrder;
            source->currPattern = modPattern;
        }
    }
}

int processSamples(int numSamples)
{
    if (audio.audioMixerOpen == true)
    {
        if (sampleCounter == 0)
        {
            if (editor.songPlaying == true)
            {
                processTick();
            }

            sampleCounter += samplesPerTick;
        }

        sampleCounter -= numSamples;
        if (sampleCounter < 0)
        {
            int tempSamples = numSamples + sampleCounter;
            sampleCounter = 0;

            return tempSamples;
        }
    }

    return numSamples;
}

void modSetPos(int order, int row)
{
    if (row != -1)
    {
        modRow = row;
        source->currRow = row;

        PBreakPosition = 0;
        PosJumpAssert = false;

        modTick = 0;
    }

    if (order != -1)
    {
        if (order < source->head.orderCount)
        {
            int i;
            for (i = 0; i < PAULA_CHANNELS; ++i)
            {
                source->channels[i].patternLoopRow = 0;
                source->channels[i].patternLoopCounter = 0;
            }

            modOrder = order;
            source->currOrder = order;

            modPattern = source->head.order[order];
            if (modPattern >= source->MAX_PATTERNS)
            {
                modPattern = source->MAX_PATTERNS - 1;
            }

            source->currPattern = modPattern;

            PBreakPosition = 0;
            PosJumpAssert = false;

            modTick = 0;
        }
    }
}

void modSetSpeed(int speed)
{
    modSpeed = speed;
    source->currSpeed = speed;

    modTick = 0;
}

void modSetTempo(int bpm)
{
    modBPM = bpm;
    source->currBPM = bpm;
    samplesPerTick = audio.tempoTimerVal / bpm;
}

void modStop(void)
{
    int i;

    mixerCutChannels();

    editor.songPlaying = false;

    for (i = 0; i < PAULA_CHANNELS; ++i)
    {
        source->channels[i].patternLoopCounter = 0;
        source->channels[i].glissandoControl = 0;
        source->channels[i].vibratoControl = 0;
        source->channels[i].tremoloControl = 0;
        source->channels[i].fineTune = 0;
        source->channels[i].invertLoopSpeed = 0;
        source->channels[i].offsetBugNotAdded = false;
    }

    tempFlags = 0;

    PattDelayTime = 0;
    PattDelayTime2 = 0;

    PBreakPosition = 0;
    PosJumpAssert = false;
}

void modPlay(void)
{
    int i;

    mixerCutChannels();

    for (i = 0; i < PAULA_CHANNELS; ++i)
    {
        source->channels[i].volume = 64;
        source->channels[i].chanIndex = (char)i;
        source->channels[i].patternLoopRow = 0;
        source->channels[i].patternLoopCounter = 0;
        source->channels[i].glissandoControl = 0;
        source->channels[i].vibratoControl = 0;
        source->channels[i].vibratoPos = 0;
        source->channels[i].tremoloControl = 0;
        source->channels[i].tremoloPos = 0;
        source->channels[i].fineTune = 0;
        source->channels[i].offsetBugNotAdded = false;
    }

    source->currRow = 0;
    modRow = 0;

    if (editor.playMode == PLAY_MODE_NORMAL)
    {
        editor.playTime = 0;
    }

    tempFlags = 0;
    modTick = 0;

    PBreakPosition = 0;
    PosJumpAssert = false;

    editor.songPlaying = true;
}

void clearSong(void)
{
    if (source != NULL)
    {
        int i;

        editor.muted[0] = false;
        editor.muted[1] = false;
        editor.muted[2] = false;
        editor.muted[3] = false;

        memset(source->head.order, 0, 128);
        source->head.orderCount = 1;
        source->head.patternCount = 1;
        source->MAX_PATTERNS = 100; // M!K!, PT2.x etc

        source->head.moduleSize = 2108;

        memset(source->head.moduleTitle, 0, 21);

        editor.currentSample = 0;
        editor.playTime = 0;

        for (i = 0; i < source->MAX_PATTERNS; ++i)
        {
            if (source->patterns[i] != NULL)
            {
                free(source->patterns[i]);
            }

            source->patterns[i] = (modnote_t *)calloc(MOD_ROWS * PAULA_CHANNELS, sizeof (modnote_t));
        }

        modSetPos(0, 0);

        modSetSpeed(6);
        modSetTempo(125);

        editor.useLEDFilter = false;
    }
}

void clearSamples(void)
{
    if (source != NULL)
    {
        int i;
        for (i = 0; i < MOD_SAMPLES; ++i)
        {
            source->samples[i].fineTune = 0;
            source->samples[i].length = 0;
            source->samples[i].loopLength = 2;
            source->samples[i].loopStart = 0;
            source->samples[i].offset = 0;
            source->samples[i].volume = 0;

            memset(source->samples[i].text, 0, 23);
        }

        editor.currentSample = 0;

        if (source->sampleData != NULL)
        {
            free(source->sampleData);
        }

        source->sampleData = (signed char *)calloc(MOD_SAMPLES, 0xFFFF);

        source->head.moduleSize -= source->head.totalSampleSize;
        source->head.totalSampleSize = 0;
    }
}

void clearAll(void)
{
    if (source != NULL)
    {
        clearSamples();
        clearSong();
    }
}

void modFree(MODULE* s)
{
    if (s != NULL)
    {
        int i;
        for (i = 0; i < 100; ++i)
        {
            if (s->patterns[i] != NULL)
            {
                free(s->patterns[i]);
            }
        }

        if (s->sampleData != NULL)
        {
            free(s->sampleData);
        }

        free(s);
    }
}
